from __future__ import annotations

import itertools
from datetime import date, datetime, time, timedelta, timezone

import dateutil
import pytest
import pytz

from ibis.common.patterns import CoercedTo
from ibis.common.temporal import (
    DateUnit,
    IntervalUnit,
    TimeUnit,
    normalize_datetime,
    normalize_timedelta,
    normalize_timezone,
)
from ibis.conftest import WINDOWS

interval_units = pytest.mark.parametrize(
    ["singular", "plural", "short"],
    [
        ("year", "years", "Y"),
        ("quarter", "quarters", "Q"),
        ("month", "months", "M"),
        ("week", "weeks", "W"),
        ("day", "days", "D"),
        ("hour", "hours", "h"),
        ("minute", "minutes", "m"),
        ("second", "seconds", "s"),
        ("millisecond", "milliseconds", "ms"),
        ("microsecond", "microseconds", "us"),
        ("nanosecond", "nanoseconds", "ns"),
    ],
)


@interval_units
def test_interval_units(singular, plural, short):
    u = IntervalUnit[singular.upper()]
    assert u.singular == singular
    assert u.plural == plural
    assert u.short == short


@interval_units
def test_interval_unit_coercions(singular, plural, short):
    u = IntervalUnit[singular.upper()]
    v = CoercedTo(IntervalUnit)
    assert v.match(singular, {}) == u
    assert v.match(plural, {}) == u
    assert v.match(short, {}) == u


@pytest.mark.parametrize(
    ("alias", "expected"),
    [
        ("HH24", "h"),
        ("J", "D"),
        ("MI", "m"),
        ("SYYYY", "Y"),
        ("YY", "Y"),
        ("YYY", "Y"),
        ("YYYY", "Y"),
    ],
)
def test_interval_unit_aliases(alias, expected):
    v = CoercedTo(IntervalUnit)
    assert v.match(alias, {}) == IntervalUnit(expected)


@pytest.mark.parametrize(
    ("value", "unit", "expected"),
    [
        (1, IntervalUnit.DAY, 1),
        (1, IntervalUnit.HOUR, 1),
        (1, IntervalUnit.MINUTE, 1),
        (1, IntervalUnit.SECOND, 1),
        (1, IntervalUnit.MILLISECOND, 1),
        (1, IntervalUnit.MICROSECOND, 1),
        (timedelta(days=1), IntervalUnit.DAY, 1),
        (timedelta(hours=1), IntervalUnit.HOUR, 1),
        (timedelta(minutes=1), IntervalUnit.MINUTE, 1),
        (timedelta(seconds=1), IntervalUnit.SECOND, 1),
        (timedelta(milliseconds=1), IntervalUnit.MILLISECOND, 1),
        (timedelta(microseconds=1), IntervalUnit.MICROSECOND, 1),
        (timedelta(seconds=1, milliseconds=100), IntervalUnit.MILLISECOND, 1100),
        (timedelta(seconds=1, milliseconds=21), IntervalUnit.MICROSECOND, 1021000),
    ],
)
def test_normalize_timedelta(value, unit, expected):
    assert normalize_timedelta(value, unit) == expected


@pytest.mark.parametrize(
    ("value", "unit"),
    [
        (timedelta(days=1), IntervalUnit.YEAR),
        (timedelta(days=1), IntervalUnit.QUARTER),
        (timedelta(days=1), IntervalUnit.MONTH),
        (timedelta(days=1), IntervalUnit.WEEK),
        (timedelta(hours=1), IntervalUnit.DAY),
        (timedelta(minutes=1), IntervalUnit.HOUR),
        (timedelta(seconds=1), IntervalUnit.MINUTE),
        (timedelta(milliseconds=1), IntervalUnit.SECOND),
        (timedelta(microseconds=1), IntervalUnit.MILLISECOND),
        (timedelta(days=1, microseconds=100), IntervalUnit.MILLISECOND),
    ],
)
def test_normalize_timedelta_invalid(value, unit):
    with pytest.raises(ValueError):
        normalize_timedelta(value, unit)


def test_interval_unit_compatibility():
    v = CoercedTo(IntervalUnit)
    for unit in itertools.chain(DateUnit, TimeUnit):
        interval = v.match(unit, {})
        assert isinstance(interval, IntervalUnit)
        assert unit.value == interval.value


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (None, None),
        (pytz.UTC, pytz.UTC),
        ("UTC", dateutil.tz.tzutc()),
        ("Europe/Budapest", dateutil.tz.gettz("Europe/Budapest")),
        (+2, timezone(timedelta(seconds=7200))),
        (-2, timezone(timedelta(seconds=-7200))),
        (dateutil.tz.tzoffset(None, 3600), timezone(timedelta(seconds=3600))),
    ],
)
def test_normalize_timezone(value, expected):
    assert normalize_timezone(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        # datetime object
        (datetime(2017, 1, 1), datetime(2017, 1, 1)),
        (datetime(2017, 1, 1, 0, 0, 0, 1), datetime(2017, 1, 1, 0, 0, 0, 1)),
        (
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=timezone.utc),
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.UTC),
        ),
        # date object
        (datetime(2017, 1, 1).date(), datetime(2017, 1, 1)),
        # datetime string
        ("2017-01-01", datetime(2017, 1, 1)),
        ("2017-01-01 00:00:00.000001", datetime(2017, 1, 1, 0, 0, 0, 1)),
        # datetime string with timezone offset
        (
            "2017-01-01 00:00:00.000001+00:00",
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.UTC),
        ),
        (
            "2017-01-01 00:00:00.000001+01:00",
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.tzoffset(None, 3600)),
        ),
        # datetime string with timezone
        (
            "2017-01-01 00:00:00.000001 UTC",
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.UTC),
        ),
        (
            "2017-01-01 00:00:00.000001 GMT",
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.UTC),
        ),
        # plain integer
        (1000, datetime(1970, 1, 1, 0, 16, 40, tzinfo=dateutil.tz.UTC)),
        # floating point
        (1000.123, datetime(1970, 1, 1, 0, 16, 40, 123000, tzinfo=dateutil.tz.UTC)),
    ],
)
def test_normalize_datetime(value, expected):
    result = normalize_datetime(value)
    assert result == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("2017-01-01", datetime(2017, 1, 1)),
        ("2017-01-01 00:00:00.000001", datetime(2017, 1, 1, 0, 0, 0, 1)),
        (
            "2017-01-01 00:00:00.000001+00:00",
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.UTC),
        ),
        (
            "2017-01-01 00:00:00.000001+01:00",
            datetime(2017, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.tzoffset(None, 3600)),
        ),
    ],
)
def test_normalize_datetime_pandas(value, expected):
    pd = pytest.importorskip("pandas")
    result = normalize_datetime(pd.Timestamp(value))
    assert result == expected


def test_normalize_datetime_with_time(mocker):
    import datetime

    class MockDate:
        @classmethod
        def today(cls) -> date:
            # this CANNOT be datetime.date, that will invoke the mock
            # constructor
            return date(2024, 7, 6)

    mocker.patch("datetime.date", new=MockDate)

    # it's necessary to use the `datetime` *module* here otherwise the patch
    # will not be used because `date` is imported before the patch is applied
    today = datetime.date.today()

    value = time(0, 0, 0, 1)
    expected = datetime.datetime.combine(today, value)

    assert normalize_datetime(value) == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        # timezone naive datetime
        (datetime(2017, 1, 1), None),
        (datetime(2017, 1, 1, 0, 0, 0, 1), None),
        # timezone aware datetime
        (datetime(2022, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.UTC), "UTC"),
        # timezone aware datetime with timezone offset
        (
            datetime(2022, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.tzoffset(None, 3600)),
            "UTC+01:00",
        ),
        # timezone aware datetime with timezone name
        (datetime(2022, 1, 1, 0, 0, 0, 1, tzinfo=dateutil.tz.gettz("CET")), "CET"),
    ],
)
def test_normalized_datetime_tzname(value, expected):
    result = normalize_datetime(value)
    assert result.tzname() == expected


def test_normalized_datetime_tzname_pandas():
    pd = pytest.importorskip("pandas")

    result = normalize_datetime(pd.Timestamp("2022-01-01 00:00:00.000001+00:00"))
    assert result.tzname() == "UTC"


def test_normalized_datetime_tzname_pandas_non_utc():
    from packaging.version import parse as vparse

    pd = pytest.importorskip("pandas")

    if vparse(pd.__version__) < vparse("2.0.0") and not WINDOWS:
        pytest.xfail(
            "tzdata is missing in pandas < 2.0.0 due to an incorrect marker "
            "in the tzdata package specification that restricts its installation "
            "to windows only"
        )

    result = normalize_datetime(pd.Timestamp("2022-01-01 00:00:00.000001+01:00"))
    assert result.tzname() == "UTC+01:00"



from __future__ import annotations

import os
import subprocess
from typing import TYPE_CHECKING, Literal

import pytest
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.expr.schema as sch
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import ServiceBackendTest

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


TRINO_USER = os.environ.get(
    "IBIS_TEST_TRINO_USER", os.environ.get("TRINO_USER", "user")
)
TRINO_PASS = os.environ.get(
    "IBIS_TEST_TRINO_PASSWORD", os.environ.get("TRINO_PASSWORD", "")
)
TRINO_HOST = os.environ.get(
    "IBIS_TEST_TRINO_HOST", os.environ.get("TRINO_HOST", "localhost")
)
TRINO_PORT = int(
    os.environ.get("IBIS_TEST_TRINO_PORT", os.environ.get("TRINO_PORT", 8080))
)


class TestConf(ServiceBackendTest):
    # trino rounds half to even for double precision and half away from zero
    # for numeric and decimal

    service_name = "minio"
    data_volume = "/bitnami/minio/data"
    returned_timestamp_unit = "s"
    supports_structs = True
    supports_map = True
    supports_tpch = True
    supports_tpcds = True
    deps = ("trino",)

    def preload(self):
        # copy files to the minio host
        super().preload()

        for suite in ["tpch", "tpcds"]:
            for path in self.data_dir.joinpath(suite).rglob("*.parquet"):
                subprocess.run(
                    [
                        "docker",
                        "compose",
                        "cp",
                        str(path),
                        f"{self.service_name}:{self.data_volume}/{suite}_{path.name}",
                    ],
                    check=False,
                )

                dirname = path.with_suffix("").name
                subprocess.run(
                    [
                        "docker",
                        "compose",
                        "exec",
                        self.service_name,
                        "mc",
                        "cp",
                        f"{self.data_volume}/{suite}_{path.name}",
                        f"data/trino/{suite}/{dirname}/",
                    ],
                    check=True,
                )

        for path in self.test_files:
            # minio doesn't allow underscores in bucket names
            dirname = path.with_suffix("").name
            # copy from minio container to trino minio host
            subprocess.run(
                [
                    "docker",
                    "compose",
                    "exec",
                    self.service_name,
                    "mc",
                    "cp",
                    f"{self.data_volume}/{path.name}",
                    f"data/trino/{dirname}/",
                ],
                check=True,
            )

    def _tpc_table(self, name: str, benchmark: Literal["h", "ds"]):
        return self.connection.table(name, database=f"hive.tpc{benchmark}")

    def _transform_tpc_sql(self, parsed, *, suite, leaves):
        def add_catalog_and_schema(node):
            if isinstance(node, sg.exp.Table) and node.name in leaves:
                catalog = "hive"
                db = f"tpc{suite}"
                return node.__class__(
                    db=db,
                    catalog=catalog,
                    **{
                        k: v for k, v in node.args.items() if k not in ("db", "catalog")
                    },
                )
            return node

        result = parsed.transform(add_catalog_and_schema)
        return result

    def _load_tpc(self, *, suite, **_) -> None:
        """Create views of data in the TPC-H catalog that ships with Trino.

        This method create relations that have column names prefixed with the
        first one (or two in the case of partsupp -> ps) character table name
        to match the DuckDB TPC-H query conventions.
        """
        suite_name = f"tpc{suite}"
        sqls = generate_tpc_tables(suite_name, data_dir=self.data_dir)
        with self.connection.begin() as con:
            con.execute(f"CREATE SCHEMA IF NOT EXISTS hive.{suite_name}")
            for stmt in sqls:
                raw_sql = stmt.sql("trino", pretty=True)
                con.execute(raw_sql)

    @property
    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath("parquet").glob("*.parquet")

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.trino.connect(
            host=TRINO_HOST,
            port=TRINO_PORT,
            user=TRINO_USER,
            auth=TRINO_PASS,
            database="memory",
            schema="default",
            **kw,
        )

    def _remap_column_names(self, table_name: str) -> dict[str, str]:
        table = self.connection.table(table_name)
        return table.rename(
            dict(zip(TEST_TABLES[table_name].names, table.schema().names))
        )

    @property
    def batting(self):
        return self._remap_column_names("batting")

    @property
    def awards_players(self):
        return self._remap_column_names("awards_players")


@pytest.fixture(scope="session")
def con(tmp_path_factory, data_dir, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


def generate_tpc_tables(suite_name, *, data_dir):
    import pyarrow.parquet as pq

    tables = {
        path.with_suffix("").name: sch.from_pyarrow_schema(
            pq.read_metadata(path).schema.to_arrow_schema()
        )
        for path in (data_dir / suite_name).rglob("*.parquet")
    }
    return (
        sge.Create(
            kind="TABLE",
            exists=True,
            this=sge.Schema(
                this=sg.table(name, db=suite_name, catalog="hive", quoted=True),
                expressions=schema.to_sqlglot("trino"),
            ),
            properties=sge.Properties(
                expressions=[
                    sge.Property(
                        this="external_location",
                        value=sge.convert(f"s3a://trino/{suite_name}/{name}"),
                    ),
                    sge.Property(this="format", value=sge.convert("PARQUET")),
                ]
            ),
        )
        for name, schema in tables.items()
    )



from __future__ import annotations

import datetime

from ibis.formats.pandas import PandasData


class TrinoPandasData(PandasData):
    @classmethod
    def convert_Interval(cls, s, dtype, pandas_dtype):
        def parse_trino_timedelta(value):
            if isinstance(value, str):
                # format is 'days hour:minute:second.millisecond'
                days, rest = value.split(" ", 1)
                hms, millis = rest.split(".", 1)
                hours, minutes, seconds = hms.split(":")
                return datetime.timedelta(
                    days=int(days),
                    hours=int(hours),
                    minutes=int(minutes),
                    seconds=int(seconds),
                    milliseconds=int(millis),
                )
            return value

        return s.map(parse_trino_timedelta, na_action="ignore")



from __future__ import annotations

import pandas.testing as tm
import pytest

import ibis
from ibis.backends.pyspark import PYSPARK_LT_35
from ibis.conftest import IS_SPARK_REMOTE

pytest.importorskip("pyspark")


@pytest.fixture
def t(con):
    return con.table("basic_table")


@pytest.fixture
def df(con):
    return con._session.table("basic_table").toPandas()


@ibis.udf.scalar.builtin
def repeat(x, n) -> str: ...


@ibis.udf.scalar.python
def py_repeat(x: str, n: int) -> str:
    return x * n


@ibis.udf.scalar.pyarrow
def pyarrow_repeat(x: str, n: int) -> str:
    return x * n


def test_builtin_udf(t, df):
    result = t.mutate(repeated=repeat(t.str_col, 2)).execute()
    expected = df.assign(repeated=df.str_col * 2)
    tm.assert_frame_equal(result, expected)


def test_python_udf(t, df):
    result = t.mutate(repeated=py_repeat(t.str_col, 2)).execute()
    expected = df.assign(repeated=df.str_col * 2)
    tm.assert_frame_equal(result, expected)


@pytest.mark.xfail(PYSPARK_LT_35, reason="pyarrow UDFs require PySpark 3.5+")
@pytest.mark.xfail(
    IS_SPARK_REMOTE,
    reason="pyarrow UDFs aren't tested with spark remote due to environment setup complexities",
)
def test_pyarrow_udf(t, df):
    result = t.mutate(repeated=pyarrow_repeat(t.str_col, 2)).execute()
    expected = df.assign(repeated=df.str_col * 2)
    tm.assert_frame_equal(result, expected)


@pytest.mark.xfail(not PYSPARK_LT_35, reason="pyarrow UDFs require PySpark 3.5+")
def test_illegal_udf_type(t):
    @ibis.udf.scalar.pyarrow
    def my_add_one(x) -> str:
        import pyarrow.compute as pac

        return pac.add(pac.binary_length(x), 1)

    expr = t.select(repeated=my_add_one(t.str_col))

    with pytest.raises(
        NotImplementedError,
        match="pyarrow UDFs are only supported in pyspark >= 3.5",
    ):
        expr.execute()



"""The Oracle backend."""

from __future__ import annotations

import contextlib
import re
import warnings
from functools import cached_property
from operator import itemgetter
from typing import TYPE_CHECKING, Any
from urllib.parse import unquote_plus

import oracledb
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanListDatabase
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import STAR, C

if TYPE_CHECKING:
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import pyarrow as pa


def metadata_row_to_type(
    *, type_mapper, type_string, precision, scale, nullable
) -> dt.DataType:
    """Convert a row from an Oracle metadata table to an Ibis type."""
    # See
    # https://docs.oracle.com/en/database/oracle/oracle-database/19/sqlrf/Data-Types.html#GUID-0BC16006-32F1-42B1-B45E-F27A494963FF
    # for details
    #
    # NUMBER(null, null) --> NUMBER(38) -> NUMBER(38, 0)
    # (null, null) --> from_string()
    if type_string == "NUMBER" and precision is None and not scale:
        typ = dt.Int64(nullable=nullable)

    # (null, 0) --> INT
    # (null, 3), (null, 6), (null, 9) --> from_string() - TIMESTAMP(3)/(6)/(9)
    elif precision is None and (scale is not None and scale == 0):
        typ = dt.Int64(nullable=nullable)

    # NUMBER(*, 0) --> INT
    # (*, 0) --> from_string() - INTERVAL DAY(3) TO SECOND(0)
    elif (
        type_string == "NUMBER"
        and precision is not None
        and (scale is not None and scale == 0)
    ):
        typ = dt.Int64(nullable=nullable)

    # NUMBER(*, > 0) --> DECIMAL
    # (*, > 0) --> from_string() - INTERVAL DAY(3) TO SECOND(2)
    elif (
        type_string == "NUMBER"
        and precision is not None
        and (scale is not None and scale > 0)
    ):
        typ = dt.Decimal(precision=precision, scale=scale, nullable=nullable)

    else:
        typ = type_mapper.from_string(type_string, nullable=nullable)
    return typ


class Backend(SQLBackend, CanListDatabase):
    name = "oracle"
    compiler = sc.oracle.compiler

    @cached_property
    def version(self):
        matched = re.search(r"(\d+)\.(\d+)\.(\d+)", self.con.version)
        return ".".join(matched.groups())

    def do_connect(
        self,
        *,
        user: str,
        password: str,
        host: str = "localhost",
        port: int = 1521,
        database: str | None = None,
        sid: str | None = None,
        service_name: str | None = None,
        dsn: str | None = None,
        **_: Any,
    ) -> None:
        """Create an Ibis client using the passed connection parameters.

        Parameters
        ----------
        user
            Username
        password
            Password
        host
            Hostname
        port
            Port
        database
            Used as an Oracle service name if provided.
        sid
            Unique name of an Oracle Instance, used to construct a DSN if
            provided.
        service_name
            Oracle service name, used to construct a DSN if provided.  Only one
            of database and service_name should be provided.
        dsn
            An Oracle Data Source Name.  If provided, overrides all other
            connection arguments except username and password.

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_ORACLE_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_ORACLE_USER", "ibis")
        >>> password = os.environ.get("IBIS_TEST_ORACLE_PASSWORD", "ibis")
        >>> database = os.environ.get("IBIS_TEST_ORACLE_DATABASE", "IBIS_TESTING")
        >>> con = ibis.oracle.connect(database=database, host=host, user=user, password=password)
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        DatabaseTable: functional_alltypes
          id              int64
          bool_col        int64
          tinyint_col     int64
          smallint_col    int64
          int_col         int64
          bigint_col      int64
          float_col       float64
          double_col      float64
          date_string_col string
          string_col      string
          timestamp_col   timestamp(3)
          year            int64
          month           int64
        """
        # SID: unique name of an INSTANCE running an oracle process (a single, identifiable machine)
        # service name: an ALIAS to one (or many) individual instances that can
        # be hotswapped without the client knowing / caring
        if dsn is not None and (
            database is not None or sid is not None or service_name is not None
        ):
            warnings.warn(
                "Oracle DSN provided, overriding additional provided connection arguments"
            )

        if service_name is not None and database is not None:
            raise exc.IbisInputError(
                "Values provided for both service_name and database. "
                "Both of these values map to an Oracle service_name, "
                "please provide only one of them."
            )

        if service_name is None and database is not None:
            service_name = database

        if dsn is None:
            dsn = oracledb.makedsn(host, port, service_name=service_name, sid=sid)

        # We set the statement cache size to 0 because Oracle will otherwise
        # attempt to reuse prepared statements even if the type of the bound variable
        # has changed.
        # This is apparently accepted behavior.
        # https://python-oracledb.readthedocs.io/en/latest/user_guide/appendix_b.html#statement-caching-in-thin-and-thick-modes
        self.con = oracledb.connect(dsn, user=user, password=password, stmtcachesize=0)

        self._post_connect()

    @util.experimental
    @classmethod
    def from_connection(cls, con: oracledb.Connection) -> Backend:
        """Create an Ibis client from an existing connection to an Oracle database.

        Parameters
        ----------
        con
            An existing connection to an Oracle database.
        """
        new_backend = cls()
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect()
        return new_backend

    def _post_connect(self) -> None:
        # turn on autocommit
        # TODO: it would be great if this worked but it doesn't seem to do the trick
        # I had to hack in the commit lines to the compiler
        # self.con.autocommit = True

        # Set to ensure decimals come back as decimals
        oracledb.defaults.fetch_decimals = True

    def _from_url(self, url: ParseResult, **kwargs):
        self.do_connect(
            user=url.username,
            password=unquote_plus(url.password) if url.password is not None else None,
            database=url.path.removeprefix("/"),
            port=url.port,
            **kwargs,
        )

        return self

    @property
    def current_catalog(self) -> str:
        with self._safe_raw_sql(sg.select(STAR).from_("global_name")) as cur:
            [(catalog,)] = cur.fetchall()
        return catalog

    @property
    def current_database(self) -> str:
        # databases correspond to users, other than that there's
        # no notion of a database inside a catalog for oracle
        with self._safe_raw_sql(sg.select("user").from_("dual")) as cur:
            [(database,)] = cur.fetchall()
        return database

    @contextlib.contextmanager
    def begin(self):
        con = self.con
        cur = con.cursor()
        try:
            yield cur
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()
        finally:
            cur.close()

    @contextlib.contextmanager
    def _safe_raw_sql(self, *args, **kwargs):
        with contextlib.closing(self.raw_sql(*args, **kwargs)) as result:
            yield result

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.name)

        con = self.con
        cursor = con.cursor()

        try:
            cursor.execute(query, **kwargs)
        except Exception:
            con.rollback()
            cursor.close()
            raise
        else:
            con.commit()
            return cursor

    def list_tables(
        self,
        like: str | None = None,
        database: tuple[str, str] | str | None = None,
    ) -> list[str]:
        """List the tables in the database.

        ::: {.callout-note}
        ## Ibis does not use the word `schema` to refer to database hierarchy.

        A collection of tables is referred to as a `database`.
        A collection of `database` is referred to as a `catalog`.

        These terms are mapped onto the corresponding features in each
        backend (where available), regardless of whether the backend itself
        uses the same terminology.
        :::

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        database
            Database to list tables from. Default behavior is to show tables in
            the current database.


        """
        if database is not None:
            table_loc = database
        else:
            table_loc = self.con.username.upper()

        table_loc = self._to_sqlglot_table(table_loc)

        # Deeply frustrating here where if we call `convert` on `table_loc`,
        # which is a sg.exp.Table, the quotes are rendered as double-quotes
        # which are invalid. With no `convert`, the same thing happens.
        # If we call `convert` on the stringified SQL output, they get reparsed
        # as literal strings and those are rendered correctly.
        conditions = C.owner.eq(sge.convert(table_loc.sql(self.name)))

        tables = (
            sg.select("table_name", "owner")
            .from_(sg.table("all_tables"))
            .distinct()
            .where(conditions)
        )
        views = (
            sg.select("view_name", "owner")
            .from_(sg.table("all_views"))
            .distinct()
            .where(conditions)
        )
        sql = tables.union(views).sql(self.name)

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        if catalog is not None:
            raise exc.UnsupportedArgumentError(
                "No cross-catalog schema access in Oracle"
            )

        query = sg.select("username").from_("all_users").order_by("username")

        with self._safe_raw_sql(query) as con:
            schemata = list(map(itemgetter(0), con))

        return self._filter_with_like(schemata, like)

    def get_schema(
        self, name: str, *, catalog: str | None = None, database: str | None = None
    ) -> sch.Schema:
        if database is None:
            database = self.con.username.upper()
        stmt = (
            sg.select(
                C.column_name,
                C.data_type,
                C.data_precision,
                C.data_scale,
                C.nullable.eq(sge.convert("Y")).as_("nullable"),
            )
            .from_(sg.table("all_tab_columns"))
            .where(
                C.table_name.eq(sge.convert(name)),
                C.owner.eq(sge.convert(database)),
            )
            .order_by(C.column_id)
        )
        with self._safe_raw_sql(stmt) as cur:
            results = cur.fetchall()

        if not results:
            raise exc.TableNotFound(name)

        type_mapper = self.compiler.type_mapper
        fields = {
            name: metadata_row_to_type(
                type_mapper=type_mapper,
                type_string=type_string,
                precision=precision,
                scale=scale,
                nullable=nullable,
            )
            for name, type_string, precision, scale, nullable in results
        }

        return sch.Schema(fields)

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Oracle.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        initial_table = sg.table(temp_name, db=database, quoted=self.compiler.quoted)
        target = sge.Schema(
            this=initial_table,
            expressions=(schema or table.schema()).to_sqlglot(self.dialect),
        )

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        # This is the same table as initial_table unless overwrite == True
        final_table = sg.table(name, db=database, quoted=self.compiler.quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                insert_stmt = sge.Insert(this=initial_table, expression=query).sql(
                    self.name
                )
                cur.execute(insert_stmt)

            if overwrite:
                self.drop_table(
                    name=final_table.name, database=final_table.db, force=True
                )
                cur.execute(
                    f"ALTER TABLE IF EXISTS {initial_table.sql(self.name)} RENAME TO {final_table.sql(self.name)}"
                )

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def drop_table(
        self,
        name: str,
        database: tuple[str, str] | str | None = None,
        force: bool = False,
    ) -> None:
        table_loc = self._to_sqlglot_table(database or None)
        catalog, db = self._to_catalog_db_tuple(table_loc)

        table = sg.table(name, db=db, catalog=catalog, quoted=self.compiler.quoted)

        with self.begin() as bind:
            # global temporary tables cannot be dropped without first truncating them
            #
            # https://stackoverflow.com/questions/32423397/force-oracle-drop-global-temp-table
            #
            # ignore DatabaseError exceptions because the table may not exist
            # because it's already been deleted
            with contextlib.suppress(oracledb.DatabaseError):
                bind.execute(f"TRUNCATE TABLE {table.sql(self.name)}")

        super().drop_table(name, database=(catalog, db), force=force)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema

        name = op.name
        quoted = self.compiler.quoted
        create_stmt = sge.Create(
            kind="TABLE",
            this=sg.exp.Schema(
                this=sg.to_identifier(name, quoted=quoted),
                expressions=schema.to_sqlglot(self.dialect),
            ),
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        ).sql(self.name)

        data = op.data.to_frame().replace(float("nan"), None)
        insert_stmt = self._build_insert_template(
            name, schema=schema, placeholder=":{i:d}"
        )
        with self.begin() as cur:
            cur.execute(create_stmt)
            for start, end in util.chunks(len(data), chunk_size=128):
                cur.executemany(
                    insert_stmt, list(data.iloc[start:end].itertuples(index=False))
                )

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        name = util.gen_name("oracle_metadata")
        dialect = self.name

        try:
            sg_expr = sg.parse_one(query, into=sg.exp.Table, dialect=dialect)
        except sg.ParseError:
            sg_expr = sg.parse_one(query, dialect=dialect)

        # If query is a table, adjust the query accordingly
        if isinstance(sg_expr, sg.exp.Table):
            sg_expr = sg.select(STAR).from_(sg_expr)

        # TODO(gforsyth): followup -- this should probably be made a default
        # transform for quoting backends
        def transformer(node):
            if isinstance(node, sg.exp.Table):
                return sg.table(node.name, quoted=True)
            elif isinstance(node, sg.exp.Column):
                return sg.column(col=node.name, quoted=True)
            return node

        sg_expr = sg_expr.transform(transformer)

        this = sg.table(name, quoted=True)
        create_view = sg.exp.Create(kind="VIEW", this=this, expression=sg_expr).sql(
            dialect
        )
        drop_view = sg.exp.Drop(kind="VIEW", this=this).sql(dialect)

        metadata_query = (
            sg.select(
                C.column_name,
                C.data_type,
                C.data_precision,
                C.data_scale,
                C.nullable.eq(sge.convert("Y")),
            )
            .from_("all_tab_columns")
            .where(C.table_name.eq(sge.convert(name)))
            .order_by(C.column_id)
            .sql(dialect)
        )

        with self.begin() as con:
            con.execute(create_view)
            try:
                results = con.execute(metadata_query).fetchall()
            finally:
                # drop the view no matter what
                con.execute(drop_view)

        schema = {}

        # TODO: hand all this off to the type mapper
        type_mapper = self.compiler.type_mapper
        for name, type_string, precision, scale, nullable in results:
            typ = metadata_row_to_type(
                type_mapper=type_mapper,
                type_string=type_string,
                precision=precision,
                scale=scale,
                nullable=nullable,
            )
            schema[name] = typ

        return sch.Schema(schema)

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        # TODO(gforsyth): this can probably be generalized a bit and put into
        # the base backend (or a mixin)
        import pandas as pd

        from ibis.backends.oracle.converter import OraclePandasData

        df = pd.DataFrame.from_records(cursor, columns=schema.names, coerce_float=True)
        return OraclePandasData.convert_table(df, schema)

    def _clean_up_tmp_table(self, name: str) -> None:
        dialect = self.dialect

        ident = sg.to_identifier(name, quoted=self.compiler.quoted)

        truncate = sge.TruncateTable(expressions=[ident]).sql(dialect)
        drop = sge.Drop(kind="TABLE", this=ident).sql(dialect)

        with self.begin() as bind:
            # global temporary tables cannot be dropped without first truncating them
            #
            # https://stackoverflow.com/questions/32423397/force-oracle-drop-global-temp-table
            #
            # ignore DatabaseError exceptions because the table may not exist
            # because it's already been deleted
            with contextlib.suppress(oracledb.DatabaseError):
                bind.execute(truncate)
            with contextlib.suppress(oracledb.DatabaseError):
                bind.execute(drop)

    _finalize_memtable = _drop_cached_table = _clean_up_tmp_table



from __future__ import annotations

import subprocess
import sys
from importlib.metadata import EntryPoint
from typing import NamedTuple

import pytest
from pytest import param

import ibis

# FIXME(kszucs): the following backends require the sqlite backend loaded
# def test_backends_are_cached():
#     assert ibis.sqlite is ibis.sqlite
#     del ibis.sqlite  # delete to force recreation
#     assert ibis.sqlite is ibis.sqlite


# def test_backends_tab_completion():
#     assert hasattr(ibis, "sqlite")
#     del ibis.sqlite  # delete to ensure not real attr
#     assert "sqlite" in dir(ibis)
#     assert ibis.sqlite is ibis.sqlite
#     assert "sqlite" in dir(ibis)  # in dir even if already created


# def test_public_backend_methods():
#     public = {m for m in dir(ibis.sqlite) if not m.startswith("_")}
#     assert public == {"connect", "compile", "has_operation", "name"}


def test_missing_backend():
    msg = "module 'ibis' has no attribute 'foo'."
    with pytest.raises(AttributeError, match=msg):
        ibis.foo  # noqa: B018


def test_multiple_backends(mocker):
    class Distribution(NamedTuple):
        entry_points: list[EntryPoint]

    entrypoints = [
        EntryPoint(
            name="foo",
            value="ibis.backends.backend1",
            group="ibis.backends",
        ),
        EntryPoint(
            name="foo",
            value="ibis.backends.backend2",
            group="ibis.backends",
        ),
    ]
    return_value = entrypoints

    mocker.patch("importlib.metadata.entry_points", return_value=return_value)

    msg = r"\d+ packages found for backend 'foo'"
    with pytest.raises(RuntimeError, match=msg):
        ibis.foo  # noqa: B018


@pytest.mark.parametrize("module", ["pandas", "pyarrow"])
def test_no_import(module):
    script = f"""
import ibis
import sys

assert "{module}" not in sys.modules
"""
    subprocess.run([sys.executable, "-c", script], check=True)


def test_ibis_na_deprecation_warning():
    with pytest.warns(
        DeprecationWarning, match="The 'ibis.NA' constant is deprecated as of v9.1"
    ):
        assert ibis.NA is ibis.null()


@pytest.mark.parametrize(
    "op, api_func",
    [
        param("desc", ibis.desc),
        param("asc", ibis.asc),
    ],
)
def test_ibis_desc_asc_default(op, api_func):
    t = ibis.table(schema={"a": "int", "b": "str"})

    expr = t.order_by(getattr(t["b"], op)())
    expr_api = t.order_by(api_func("b"))

    assert expr.op() == expr_api.op()


@pytest.mark.parametrize(
    "op, api_func, nulls_first",
    [
        param("desc", ibis.desc, True),
        param("asc", ibis.asc, True),
        param("desc", ibis.desc, False),
        param("asc", ibis.asc, False),
    ],
)
def test_ibis_desc_asc(op, api_func, nulls_first):
    t = ibis.table(schema={"a": "int", "b": "str"})

    expr = t.order_by(getattr(t["b"], op)(nulls_first=nulls_first))
    expr_api = t.order_by(api_func("b", nulls_first=nulls_first))

    assert expr.op() == expr_api.op()



from __future__ import annotations

import os
import shutil
import tempfile

import pandas.testing as tm
import pytest
from pytest import param

import ibis.expr.datatypes as dt
from ibis import udf


@udf.scalar.builtin
def soundex(s: str) -> str: ...


@udf.scalar.builtin
def jarowinkler_similarity(a: str, b: str) -> float: ...


# TODO: allow multiple signatures
@udf.scalar.builtin(name="compress")
def compress_str(data: str, method: str) -> bytes: ...


@udf.scalar.builtin(name="compress")
def compress_bytes(data: bytes, method: str) -> bytes: ...


@pytest.mark.parametrize(
    ("func", "args"),
    [
        param(soundex, ("snow",), id="soundex"),
        param(jarowinkler_similarity, ("snow", "show"), id="jarowinkler_similarity"),
    ],
)
def test_builtin_scalar_udf(con, func, args):
    expr = func(*args)

    query = f"SELECT {func.__name__}({', '.join(map(repr, args))})"
    with con._safe_raw_sql(query) as cur:
        [(expected,)] = cur.fetchall()

    assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("func", "pyargs", "snowargs"),
    [
        param(compress_str, ("snow", "zstd"), ("'snow'", "'zstd'"), id="str"),
        param(compress_bytes, (b"snow", "zstd"), ("'snow'", "'zstd'"), id="bytes"),
    ],
)
def test_compress(con, func, pyargs, snowargs):
    expr = func(*pyargs)

    query = f"SELECT compress({', '.join(snowargs)})"
    with con._safe_raw_sql(query) as cur:
        [(expected,)] = cur.fetchall()

    assert con.execute(expr) == expected


@udf.agg.builtin
def minhash(x, y) -> dt.json: ...


@udf.agg.builtin
def approximate_jaccard_index(a) -> float: ...


def test_builtin_agg_udf(con):
    ft = con.tables.FUNCTIONAL_ALLTYPES.limit(2)
    ft = ft.select(mh=minhash(100, ft.string_col).over(group_by=ft.date_string_col))
    expr = ft.agg(aji=approximate_jaccard_index(ft.mh))

    result = expr.execute()
    query = """
    SELECT approximate_jaccard_index("mh") AS "aji"
    FROM (
        SELECT minhash(100, "string_col") OVER (PARTITION BY "date_string_col") AS "mh"
        FROM (
            SELECT * FROM "FUNCTIONAL_ALLTYPES" LIMIT 2
        )
    )
    """
    with con._safe_raw_sql(query) as cur:
        expected = cur.fetch_pandas_all()

    tm.assert_frame_equal(result, expected)


def test_xgboost_model(con):
    from ibis import _

    @udf.scalar.pandas(
        packages=("joblib", "xgboost"), imports=("@MODELS/model.joblib",)
    )
    def predict_price(
        carat_scaled: float, cut_encoded: int, color_encoded: int, clarity_encoded: int
    ) -> int:
        import sys
        from pathlib import Path

        import joblib
        import pandas as pd

        import_dir = Path(sys._xoptions.get("snowflake_import_directory"))
        assert import_dir.exists(), import_dir

        model_path = import_dir / "model.joblib"
        assert model_path.exists(), model_path

        model = joblib.load(model_path)

        df = pd.concat(
            [carat_scaled, cut_encoded, color_encoded, clarity_encoded], axis=1
        )
        df.columns = ["CARAT_SCALED", "CUT_ENCODED", "COLOR_ENCODED", "CLARITY_ENCODED"]
        return model.predict(df)

    diamonds = con.tables.DIAMONDS
    expr = diamonds.mutate(
        predicted_price=predict_price(
            (_.carat - _.carat.mean()) / _.carat.std(),
            _.cut.cases(
                *(
                    (c, i)
                    for i, c in enumerate(
                        ("Fair", "Good", "Very Good", "Premium", "Ideal"), start=1
                    )
                )
            ),
            _.color.cases(*((c, i) for i, c in enumerate("DEFGHIJ", start=1))),
            _.clarity.cases(
                *(
                    (c, i)
                    for i, c in enumerate(
                        ("I1", "IF", "SI1", "SI2", "VS1", "VS2", "VVS1", "VVS2"),
                        start=1,
                    )
                )
            ),
        )
    )

    df = expr.execute()

    assert not df.empty
    assert "predicted_price" in df.columns
    assert len(df) == diamonds.count().execute()


def touch_package(pkgpath):
    # nix timestamps every file to the UNIX epoch for reproducibility
    # so we modify the utime of the _copied_ code since snowflake has
    # some annoying checks for zipping files that do not allow files
    # older than 1980 ¯\_(ツ)_/¯
    os.utime(pkgpath, None)
    for root, dirs, files in os.walk(pkgpath):
        for path in dirs + files:
            os.utime(os.path.join(root, path), None)


def add_packages(d, session):
    import parsy
    import pyarrow_hotfix
    import rich
    import sqlglot

    import ibis

    for module in (rich, parsy, sqlglot, pyarrow_hotfix):
        pkgname = module.__name__
        pkgpath = os.path.join(d, pkgname)
        shutil.copytree(os.path.dirname(module.__file__), pkgpath)
        touch_package(pkgpath)
        session.add_import(pkgname, import_path=pkgname)

    # no need to touch the package because we're using the local version
    shutil.copytree(os.path.dirname(ibis.__file__), "ibis")
    session.add_import("ibis", import_path="ibis")


@pytest.fixture
def snowpark_session():
    if not os.environ.get("SNOWFLAKE_SNOWPARK"):
        pytest.skip("SNOWFLAKE_SNOWPARK is not set")
    else:
        sp = pytest.importorskip("snowflake.snowpark")

        if connection_name := os.environ.get("SNOWFLAKE_DEFAULT_CONNECTION_NAME"):
            builder = sp.Session.builder.config("connection_name", connection_name)
        else:
            builder = sp.Session.builder.configs(
                {
                    "user": os.environ["SNOWFLAKE_USER"],
                    "account": os.environ["SNOWFLAKE_ACCOUNT"],
                    "password": os.environ["SNOWFLAKE_PASSWORD"],
                    "warehouse": os.environ["SNOWFLAKE_WAREHOUSE"],
                    "database": os.environ["SNOWFLAKE_DATABASE"],
                    "schema": os.environ["SNOWFLAKE_SCHEMA"],
                }
            )

        session = builder.create()
        session.custom_package_usage_config["enabled"] = True

        pwd = os.getcwd()

        with tempfile.TemporaryDirectory() as d:
            os.chdir(d)

            try:
                add_packages(d, session)
                yield session
            finally:
                os.chdir(pwd)
                session.clear_imports()


@pytest.mark.parametrize("execute_as", ["owner", "caller"])
def test_ibis_inside_snowpark(snowpark_session, execute_as):
    import snowflake.snowpark as sp

    def ibis_sproc(session):
        import ibis.backends.snowflake

        con = ibis.backends.snowflake.Backend.from_connection(session)

        expr = (
            con.tables.functional_alltypes.group_by("string_col")
            .agg(n=lambda t: t.count())
            .order_by("string_col")
        )

        return session.sql(ibis.to_sql(expr))

    expected = (
        snowpark_session.table('"functional_alltypes"')
        .group_by('"string_col"')
        .count()
        .rename("COUNT", '"n"')
        .order_by('"string_col"')
        .to_pandas()
    )

    local_result = ibis_sproc(snowpark_session).to_pandas()

    tm.assert_frame_equal(local_result, expected)

    name = ibis_sproc.__name__

    snowpark_session.sproc.register(
        ibis_sproc,
        name=name,
        execute_as=execute_as,
        imports=["parsy", "rich", "sqlglot", "pyarrow_hotfix", "ibis"],
        # empty struct here tells Snowflake to infer the return type from the
        # return value of the function, which is required to be a Snowpark
        # table in that case
        return_type=sp.types.StructType(),
        packages=[
            "snowflake-snowpark-python",
            "toolz",
            "atpublic",
            "pyarrow",
            "pandas",
            "numpy",
        ],
        replace=True,
    )

    remote_result = snowpark_session.call(name).to_pandas()

    tm.assert_frame_equal(remote_result, local_result)



from __future__ import annotations

import pytest

import ibis


@pytest.mark.xfail_version(pyspark=["pyspark<3.4"], reason="no catalog support")
def test_catalog_db_args(con):
    t = ibis.memtable({"epoch": [1712848119, 1712848121, 1712848155]})

    assert con.current_catalog == "spark_catalog"
    assert con.current_database == "ibis_testing"
    con.create_database("toot", catalog="local")

    # create a table in specified catalog and db
    con.create_table("t2", database=("local", "toot"), obj=t, overwrite=True)
    con.create_table("t3", database=("spark_catalog", "default"), obj=t, overwrite=True)

    assert con.current_database == "ibis_testing"

    assert "t2" not in con.list_tables()
    assert "t2" in con.list_tables(database="local.toot")
    assert "t2" in con.list_tables(database=("local", "toot"))

    assert "t3" not in con.list_tables()
    assert "t3" in con.list_tables(database="default")
    assert "t3" in con.list_tables(database="spark_catalog.default")

    con.drop_table("t2", database="local.toot")
    con.drop_table("t3", database="spark_catalog.default")

    assert "t2" not in con.list_tables(database="local.toot")
    assert "t3" not in con.list_tables(database="spark_catalog.default")

    con.drop_database("toot", catalog="local")

    assert con.current_catalog == "spark_catalog"
    assert con.current_database == "ibis_testing"


def test_create_table_no_catalog(con):
    t = ibis.memtable({"epoch": [1712848119, 1712848121, 1712848155]})

    assert con.current_database != "default"
    # create a table in specified catalog and db
    con.create_table("t2", database=("default"), obj=t, overwrite=True)

    assert "t2" not in con.list_tables()
    assert "t2" in con.list_tables(database="default")
    assert "t2" in con.list_tables(database=("default"))

    con.drop_table("t2", database="default")

    assert "t2" not in con.list_tables(database="default")
    assert con.current_database != "default"



import ibis


customer = ibis.table(
    name="customer",
    schema={
        "c_custkey": "int64",
        "c_name": "string",
        "c_address": "string",
        "c_nationkey": "int16",
        "c_phone": "string",
        "c_acctbal": "decimal",
        "c_mktsegment": "string",
        "c_comment": "string",
    },
)
lit = ibis.literal(True)
orders = ibis.table(
    name="orders",
    schema={
        "o_orderkey": "int64",
        "o_custkey": "int64",
        "o_orderstatus": "string",
        "o_totalprice": "decimal(12, 2)",
        "o_orderdate": "date",
        "o_orderpriority": "string",
        "o_clerk": "string",
        "o_shippriority": "int32",
        "o_comment": "string",
    },
)
lineitem = ibis.table(
    name="lineitem",
    schema={
        "l_orderkey": "int32",
        "l_partkey": "int32",
        "l_suppkey": "int32",
        "l_linenumber": "int32",
        "l_quantity": "decimal(15, 2)",
        "l_extendedprice": "decimal(15, 2)",
        "l_discount": "decimal(15, 2)",
        "l_tax": "decimal(15, 2)",
        "l_returnflag": "string",
        "l_linestatus": "string",
        "l_shipdate": "date",
        "l_commitdate": "date",
        "l_receiptdate": "date",
        "l_shipinstruct": "string",
        "l_shipmode": "string",
        "l_comment": "string",
    },
)
cast = ibis.literal("1995-03-15").cast("date")
joinchain = (
    customer.inner_join(
        orders,
        [(customer.c_custkey == orders.o_custkey), lit, (orders.o_orderdate < cast)],
    )
    .inner_join(
        lineitem,
        [(orders.o_orderkey == lineitem.l_orderkey), lit, (lineitem.l_shipdate > cast)],
    )
    .select(
        customer.c_custkey,
        customer.c_name,
        customer.c_address,
        customer.c_nationkey,
        customer.c_phone,
        customer.c_acctbal,
        customer.c_mktsegment,
        customer.c_comment,
        orders.o_orderkey,
        orders.o_custkey,
        orders.o_orderstatus,
        orders.o_totalprice,
        orders.o_orderdate,
        orders.o_orderpriority,
        orders.o_clerk,
        orders.o_shippriority,
        orders.o_comment,
        lineitem.l_orderkey,
        lineitem.l_partkey,
        lineitem.l_suppkey,
        lineitem.l_linenumber,
        lineitem.l_quantity,
        lineitem.l_extendedprice,
        lineitem.l_discount,
        lineitem.l_tax,
        lineitem.l_returnflag,
        lineitem.l_linestatus,
        lineitem.l_shipdate,
        lineitem.l_commitdate,
        lineitem.l_receiptdate,
        lineitem.l_shipinstruct,
        lineitem.l_shipmode,
        lineitem.l_comment,
    )
)
f = joinchain.filter((joinchain.c_mktsegment == "BUILDING"))
agg = f.aggregate(
    [(f.l_extendedprice * ((1 - f.l_discount))).sum().name("revenue")],
    by=[f.l_orderkey, f.o_orderdate, f.o_shippriority],
)
s = agg.order_by(agg.revenue.desc(), agg.o_orderdate.asc(nulls_first=True))

result = s.select(s.l_orderkey, s.revenue, s.o_orderdate, s.o_shippriority).limit(10)



from __future__ import annotations

import datetime
from collections import OrderedDict

import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis import _
from ibis.backends.tests.errors import OracleDatabaseError, PsycoPg2InternalError

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")


@pytest.mark.parametrize(
    ("column", "raw_value"),
    [
        ("double_col", 0.0),
        ("double_col", 10.1),
        ("float_col", 1.1),
        ("float_col", 2.2),
    ],
)
def test_floating_scalar_parameter(backend, alltypes, df, column, raw_value):
    value = ibis.param(dt.double)
    expr = (alltypes[column] + value).name("tmp")
    expected = df[column] + raw_value
    result = expr.execute(params={value: raw_value})
    expected = backend.default_series_rename(expected)
    backend.assert_series_equal(result, expected, check_dtype=False)


@pytest.mark.parametrize(
    ("start_string", "end_string"),
    [("2009-03-01", "2010-07-03"), ("2014-12-01", "2017-01-05")],
)
@pytest.mark.notimpl(["trino", "druid"])
@pytest.mark.notimpl(["oracle"], raises=OracleDatabaseError)
def test_date_scalar_parameter(backend, alltypes, start_string, end_string):
    start, end = ibis.param(dt.date), ibis.param(dt.date)

    col = alltypes.timestamp_col.date()
    expr = col.between(start, end).name("output")
    expected_expr = col.between(start_string, end_string).name("output")

    result = expr.execute(params={start: start_string, end: end_string})
    expected = expected_expr.execute()

    backend.assert_series_equal(result, expected)


def test_timestamp_accepts_date_literals(alltypes):
    date_string = "2009-03-01"
    param = ibis.param(dt.timestamp)
    expr = alltypes.mutate(param=param)
    params = {param: date_string}
    assert expr.compile(params=params) is not None


@pytest.mark.notimpl(["impala", "druid", "oracle", "exasol"])
@pytest.mark.never(
    ["mysql", "sqlite", "mssql"], reason="backend will never implement array types"
)
def test_scalar_param_array(con):
    value = [1, 2, 3]
    param = ibis.param(dt.Array(dt.int64))
    result = con.execute(param.length().name("tmp"), params={param: value})
    assert result == len(value)


@pytest.mark.notimpl(
    [
        "datafusion",
        "impala",
        "postgres",
        "risingwave",
        "druid",
        "oracle",
        "exasol",
    ]
)
@pytest.mark.never(
    ["mysql", "sqlite", "mssql"],
    reason="mysql and sqlite will never implement struct types",
)
def test_scalar_param_struct(con):
    value = dict(a=1, b="abc", c=3.0)
    param = ibis.param("struct<a: int64, b: string, c: float64>")
    result = con.execute(param["a"], params={param: value})
    assert result == value["a"]


@pytest.mark.notimpl(["datafusion", "impala", "polars", "druid", "oracle", "exasol"])
@pytest.mark.never(
    ["mysql", "sqlite", "mssql"],
    reason="mysql and sqlite will never implement map types",
)
@pytest.mark.notyet(["bigquery"])
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="function make_date(integer, integer, integer) does not exist",
)
def test_scalar_param_map(con):
    value = {"a": "ghi", "b": "def", "c": "abc"}
    param = ibis.param(dt.Map(dt.string, dt.string))
    result = con.execute(param["b"], params={param: value})
    assert result == value["b"]


@pytest.mark.parametrize(
    ("value", "dtype", "col"),
    [
        param("0", "string", "string_col", id="string"),
        param(0, "int64", "int_col", id="int"),
        param(0.0, "float64", "double_col", id="double"),
        param(
            True,
            "bool",
            "bool_col",
            id="bool",
            marks=[pytest.mark.notimpl(["druid"])],
        ),
        param(
            "2009-01-20 01:02:03", "timestamp", "timestamp_col", id="string_timestamp"
        ),
        param(
            datetime.date(2009, 1, 20),
            "timestamp",
            "timestamp_col",
            id="date_timestamp",
        ),
        param(
            datetime.datetime(2009, 1, 20, 1, 2, 3),
            "timestamp",
            "timestamp_col",
            id="datetime_timestamp",
        ),
    ],
)
def test_scalar_param(backend, alltypes, df, value, dtype, col):
    param = ibis.param(dtype)
    expr = alltypes.filter([_[col] == param])

    result = (
        expr.execute(params={param: value}).sort_values("id").reset_index(drop=True)
    )
    expected = df.loc[df[col] == value].sort_values("id").reset_index(drop=True)
    backend.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "value",
    ["2009-01-20", datetime.date(2009, 1, 20), datetime.datetime(2009, 1, 20)],
    ids=["string", "date", "datetime"],
)
@pytest.mark.notimpl(["druid"])
def test_scalar_param_date(backend, alltypes, value):
    param = ibis.param("date")
    ds_col = alltypes.date_string_col
    month = ds_col[:2]
    day = ds_col[3:5]
    year = "20" + ds_col[6:8]
    date_col = (year + "-" + month + "-" + day).cast(param.type())

    base = alltypes.mutate(date_col=date_col)
    expr = (
        alltypes.mutate(date_col=date_col)
        .filter(lambda t: t.date_col == param)
        .drop("date_col")
    )

    result = (
        expr.execute(params={param: value}).sort_values("id").reset_index(drop=True)
    )
    df = base.execute()
    expected = (
        df.loc[
            pd.to_datetime(df.date_col).dt.normalize().dt.date
            == pd.Timestamp(value).normalize().date()
        ]
        .sort_values("id")
        .reset_index(drop=True)
        .drop(columns=["date_col"])
    )
    backend.assert_frame_equal(result, expected)


@pytest.mark.notyet(["flink", "mysql"], reason="no struct support")
@pytest.mark.notimpl(
    [
        "postgres",
        "risingwave",
        "datafusion",
        "clickhouse",
        "sqlite",
        "impala",
        "oracle",
        "mssql",
        "druid",
        "exasol",
    ]
)
def test_scalar_param_nested(con):
    param = ibis.param("struct<x: array<struct<y: array<double>>>>")
    value = OrderedDict([("x", [OrderedDict([("y", [1.0, 2.0, 3.0])])])])
    result = con.execute(param, {param: value})
    assert pytest.approx(result["x"][0]["y"]) == np.array([1.0, 2.0, 3.0])



import ibis


lit = ibis.timestamp("2018-01-01 00:00:00")
s = ibis.table(name="s", schema={"b": "string"})
t = ibis.table(name="t", schema={"a": "int64", "b": "string", "c": "timestamp"})
f = t.filter((t.c == lit))
dropcolumns = f.select(f.a, f.b, f.c.name("C")).drop("C")
joinchain = (
    dropcolumns.select(dropcolumns.a, dropcolumns.b, lit.name("the_date"))
    .inner_join(
        s,
        (
            dropcolumns.select(dropcolumns.a, dropcolumns.b, lit.name("the_date")).b
            == s.b
        ),
    )
    .select(dropcolumns.select(dropcolumns.a, dropcolumns.b, lit.name("the_date")).a)
)

result = joinchain.filter((joinchain.a < 1.0))



from __future__ import annotations

import hypothesis as h
import hypothesis.strategies as st
import pytest

import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
import ibis.expr.types as ir
import ibis.tests.strategies as its


@h.given(its.null_dtype)
def test_null_dtype(dtype):
    assert isinstance(dtype, dt.Null)
    assert dtype.is_null() is True
    assert dtype.nullable is True


@h.given(its.boolean_dtype())
def test_boolean_dtype(dtype):
    assert isinstance(dtype, dt.Boolean)
    assert dtype.is_boolean() is True


@h.given(its.signed_integer_dtypes())
def test_signed_integer_dtype(dtype):
    assert isinstance(dtype, dt.SignedInteger)
    assert dtype.is_integer() is True


@h.given(its.unsigned_integer_dtypes())
def test_unsigned_integer_dtype(dtype):
    assert isinstance(dtype, dt.UnsignedInteger)
    assert dtype.is_integer() is True


@h.given(its.floating_dtypes())
def test_floating_dtype(dtype):
    assert isinstance(dtype, dt.Floating)
    assert dtype.is_floating() is True


@h.given(its.numeric_dtypes())
def test_numeric_dtype(dtype):
    assert isinstance(dtype, dt.Numeric)
    assert dtype.is_numeric() is True


@h.given(its.numeric_dtypes(nullable=st.just(True)))
def test_numeric_dtypes_nullable(dtype):
    assert dtype.nullable is True
    assert dtype.is_numeric() is True


@h.given(its.numeric_dtypes(nullable=st.just(False)))
def test_numeric_dtypes_non_nullable(dtype):
    assert dtype.nullable is False
    assert dtype.is_numeric() is True


@h.given(its.timestamp_dtype())
def test_timestamp_dtype(dtype):
    assert isinstance(dtype, dt.Timestamp)
    assert isinstance(dtype.timezone, (type(None), str))
    assert dtype.is_timestamp() is True


@h.given(its.interval_dtype())
def test_interval_dtype(dtype):
    assert isinstance(dtype, dt.Interval)
    assert dtype.is_interval() is True


@h.given(its.temporal_dtypes())
def test_temporal_dtype(dtype):
    assert isinstance(dtype, dt.Temporal)
    assert dtype.is_temporal() is True


@h.given(its.primitive_dtypes())
def test_primitive_dtype(dtype):
    assert isinstance(dtype, dt.Primitive)
    assert dtype.is_primitive() is True


@h.given(its.geospatial_dtypes())
def test_geospatial_dtype(dtype):
    assert isinstance(dtype, dt.GeoSpatial)
    assert dtype.is_geospatial() is True


@h.given(its.array_dtypes(its.primitive_dtypes()))
def test_array_dtype(dtype):
    assert isinstance(dtype, dt.Array)
    assert isinstance(dtype.value_type, dt.Primitive)
    assert dtype.is_array() is True


@h.given(its.array_dtypes(its.array_dtypes(its.primitive_dtypes())))
def test_array_array_dtype(dtype):
    assert isinstance(dtype, dt.Array)
    assert isinstance(dtype.value_type, dt.Array)
    assert isinstance(dtype.value_type.value_type, dt.Primitive)


@h.given(its.map_dtypes(its.primitive_dtypes(), its.boolean_dtype()))
def test_map_dtype(dtype):
    assert isinstance(dtype, dt.Map)
    assert isinstance(dtype.key_type, dt.Primitive)
    assert isinstance(dtype.value_type, dt.Boolean)
    assert dtype.is_map() is True


@h.given(its.struct_dtypes())
def test_struct_dtype(dtype):
    assert isinstance(dtype, dt.Struct)
    assert all(t.is_primitive() for t in dtype.types)
    assert dtype.is_struct() is True


@h.given(its.struct_dtypes(its.variadic_dtypes()))
def test_struct_variadic_dtype(dtype):
    assert isinstance(dtype, dt.Struct)
    assert all(t.is_variadic() for t in dtype.types)
    assert dtype.is_struct() is True


@h.given(its.variadic_dtypes())
def test_variadic_dtype(dtype):
    assert isinstance(dtype, dt.Variadic)
    assert dtype.is_variadic() is True


@h.given(its.all_dtypes())
def test_all_dtypes(dtype):
    assert isinstance(dtype, dt.DataType)


@h.given(its.schema())
def test_schema(schema):
    assert isinstance(schema, sch.Schema)
    assert all(t.is_primitive() for t in schema.types)
    assert all(isinstance(n, str) for n in schema.names)
    assert len(set(schema.names)) == len(schema.names)


@h.given(its.schema(its.array_dtypes(its.numeric_dtypes())))
def test_schema_array_dtype(schema):
    assert isinstance(schema, sch.Schema)
    assert all(t.is_array() for t in schema.types)
    assert all(isinstance(n, str) for n in schema.names)


@h.given(its.primitive_dtypes())
def test_primitive_dtypes_to_pandas(dtype):
    np = pytest.importorskip("numpy")
    assert isinstance(dtype.to_pandas(), np.dtype)


@h.given(its.schema())
def test_schema_to_pandas(schema):
    pytest.importorskip("pandas")
    pandas_schema = schema.to_pandas()
    assert len(pandas_schema) == len(schema)


@h.given(its.memtable(its.schema(its.integer_dtypes(), max_size=5)))
def test_memtable(memtable):
    assert isinstance(memtable, ir.Table)
    assert isinstance(memtable.schema(), sch.Schema)


# TODO(kszucs): we enforce field name uniqueness in the schema, but we don't for Struct datatype



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from pandas._libs.parsers import STR_NA_VALUES

import ibis
import ibis.expr.types as ir
from ibis import _

if TYPE_CHECKING:
    from collections.abc import Iterable

# STR_NA_VALUES is a set of strings to interpret as nan, such as
# {'null', '-1.#QNAN', '#N/A', '', '#N/A N/A', 'n/a', 'nan', 'NaN', ...}
# Add our own strings here.
NAN_LIKE = STR_NA_VALUES | {
    "anonymous",
    "anon",
    "need",
    "none",
    "not known",
    "not provided",
    "not reported",
    "not specified",
    "pending",
    "unknown",
    "unidentified",
}
NAN_LIKE = {s.lower() for s in NAN_LIKE}


def contains_alphanum(col: ir.StringValue) -> ir.BooleanValue:
    return col.lower().re_search(r".*[a-z0-9].*")


def is_nan_like(col: ir.Value) -> ir.BooleanValue:
    if not col.type().is_string():
        return col.isnull()
    result = col.isnull()
    result |= col.lower().isin(NAN_LIKE).fill_null(False)
    result |= ~contains_alphanum(col).fill_null(False)
    return result


def normalize_nulls(data: ir.StringValue) -> ir.StringValue:
    """Fill any nan-likes with NULL."""
    return is_nan_like(data).ifelse(ibis.null(), data)


def starts_with(a: ir.StringValue, b: ir.StringValue) -> ir.BooleanValue:
    """Apply a.startswith(b) to each pair of values in a and b."""
    return a.like(b + "%")


def strip(s: ir.StringValue, chars: Iterable[str]) -> ir.StringValue:
    """Remove leading and trailing characters from a string."""

    def _strip_pattern(chars: Iterable[str]) -> str:
        # If chars contains a "(" or some other regex keyword, we want
        # it to match a literal "("
        chars = (re.escape(c) for c in chars)
        char_pattern = "|".join(chars)
        char_pattern = f"({char_pattern})+"
        return char_pattern

    s = s.cast("string")
    return s.re_replace(
        "^" + _strip_pattern(chars) + "|" + _strip_pattern(chars) + "$",
        "",
    )


def fix_nicknames(t: ir.Table) -> ir.Table:
    """Splits "Nicholas (Nick)" into "Nicholas" and "Nick".

    Also deals with the form "Nicholas 'Nick'" etc.
    """
    first, nick = _parse_nicknames(t["first_name"], t["nickname"])
    return t.mutate(first_name=first, nickname=nick)


def _parse_nicknames(
    first: ir.StringValue, nick: ir.StringValue
) -> tuple[ir.StringValue, ir.StringValue]:
    enclosures = [
        r"'.*'",
        r'".*"',
        r"\(.*\)",
    ]
    enclosure_pattern = "|".join(enclosures)
    pattern = "(.*)(" + enclosure_pattern + ")"
    # 1 based
    before = first.re_extract(pattern, 1)
    inside = first.re_extract(pattern, 2)
    inside = strip(inside, "'\"()")
    before = norm_whitespace(before)
    inside = norm_whitespace(inside)

    # names_before = spacy.names_in_texts(before)
    # names_inside = spacy.names_in_texts(inside)
    # nb = names_before.apply(lambda x: x[0] if len(x) > 0 else pd.NA)
    # ni = names_inside.apply(lambda x: x[0] if len(x) > 0 else pd.NA)
    # fixable = nb.isna() & is_single_name_series(inside)
    # nb[fixable] = inside[fixable]

    needs_fix = first.notnull() & nick.isnull()
    before_good = before.split(" ").length() == 1
    inside_good = inside.split(" ").length() == 1

    todo = needs_fix & before_good & inside_good
    return todo.ifelse(before, first), todo.ifelse(inside, nick)


def to_alphanum(s: ir.StringValue, fill: str = " ") -> ir.StringValue:
    s = to_ascii(s)
    s = s.re_replace(r"[^A-Za-z0-9]+", fill)
    s = norm_whitespace(s)
    return s


def norm_whitespace(s: ir.StringValue) -> ir.StringValue:
    """Convert all whitespace to a single space, strip leading/trailing spaces."""
    s = s.cast("string")
    s = s.re_replace(r"\s+", " ")
    s = s.strip()
    s = s.nullif("")
    return s


def to_ascii(s: ir.StringValue) -> ir.StringValue:
    """Remove any non-ascii characters."""
    # return norm_whitespace(s.fill_null("").apply(unidecode).astype(s.dtype))
    # We don't have access to the unidecode function, so just strip out
    # non-ascii characters
    s = s.cast("string")
    return s.re_replace(r"[^\x00-\x7F]+", "")


def num_tokens(s: ir.StringValue) -> ir.IntegerValue:
    """Count the number of tokens, separated by spaces, in a string."""
    s = s.cast("string")
    s = s.re_replace(r"\s+", " ")
    s = s.strip()
    s = s.nullif("")
    return s.split(" ").length().fill_null(0)


NAME_COLUMNS = [
    "prefix",
    "first_name",
    "middle_name",
    "last_name",
    "suffix",
    "nickname",
]


# It would be possible to do something more sophisticated here to determine
# which of first or last to clear. But I wanted to keep this so it only operates
# on a per-row basis. This prevents us from looking atir.Table-wide statistics
# on the names to determine if a name is more likely a first or last name,
# because then the result of one reow would depend on the rest of the rows.
# In the future we could use an external lookupir.Table of likelihoods.
def drop_first_when_same_as_last(t: ir.Table) -> ir.Table:
    """Where the f and l are each one token and the same, set the first to null."""

    def _norm(s: ir.StringColumn) -> ir.StringColumn:
        s = to_alphanum(s)
        s = s.lower()
        s = s.strip()
        return s

    fn = _norm(t.first_name)
    ln = _norm(t.last_name)
    todo = ibis.and_(
        fn == ln,
        num_tokens(fn) == 1,
        num_tokens(ln) == 1,
    )
    return t.mutate(first_name=todo.ifelse(ibis.null(), t.first_name))


def fix_duplicate_appearances(t: ir.Table) -> ir.Table:
    """If a field appears as a token in another field, remove it from the other field.

    for example:
    >>> names = pd.DataFrame(
    ...     [
    ...         {"first_name": "J ang", "middle_name": "A Ang", "last_name": "ANG"},
    ...         {"first_name": "smith john", "middle_name": "B", "last_name": "SMITH"},
    ...     ]
    ... )
    >>> scrub_duplicate_appearances(names, "last_name")
    pd.DataFrame(
        [
            {'first_name': 'J', 'middle_name': 'A', 'last_name': 'ANG'},
            {'first_name': 'john', 'middle_name': 'B', 'last_name': 'SMITH'},
        ]
    )
    """
    SEP = " "
    token_cols = {c + "_tokens": _[c].split(SEP) for c in NAME_COLUMNS}
    t2 = t.mutate(**token_cols)

    def single_or_null(col):
        # If a column is a single token, choose it. Otherwise null.
        is_single = _[col + "_tokens"].length() == 1
        return is_single.ifelse(_[col].upper(), ibis.null())

    single_or_null_cols = {c + "_single": single_or_null(c) for c in NAME_COLUMNS}
    t3 = t2.mutate(**single_or_null_cols)

    def singles_except(col):
        # merge all columns into a single column of type array
        # e.g.
        # prefix_single first_single middle_single last_single
        #  a               b           null            d
        # singles_except('prefix') -> ['b', null, 'd']
        all_but_c = (c + "_single" for c in NAME_COLUMNS if c != col)
        return ibis.array([t3[c] for c in all_but_c])

    se = {"singles_except_" + c: singles_except(c) for c in NAME_COLUMNS}
    t4 = t3.mutate(**se)

    def filter_tokens(t, col):
        # can't use Deferreds until this bug fixed:
        # https://github.com/ibis-project/ibis/issues/7626
        return t[col + "_tokens"].filter(
            lambda token: ~t[f"singles_except_{col}"].contains(token.upper())
        )

    tokens_filtered = {
        c + "_tokens_filtered": filter_tokens(t4, c) for c in NAME_COLUMNS
    }
    t5 = t4.mutate(**tokens_filtered)

    final_changes = {c: _[c + "_tokens_filtered"].join(SEP) for c in NAME_COLUMNS}
    return t5.mutate(**final_changes).select(*t.columns)


def choose_longer(s1: ir.StringColumn, s2: ir.StringColumn) -> ir.StringColumn:
    l1 = s1.length().fill_null(0)
    l2 = s2.length().fill_null(0)
    return (l1 > l2).ifelse(s1, s2)


def parse_middle(
    first: ir.StringColumn, middle: ir.StringColumn, last: ir.StringColumn
) -> tuple[ir.StringColumn, ir.StringColumn, ir.StringColumn]:
    """Extract the middle name from the first or last name field.

    Four kinds of parsings
    A jones     -> Parse
    AJ          -> Inconclusive, could be first name.
    A J         -> Inconclusive, could be first name.
    Alice J     -> Parse
    Alice Jones -> Inconclusive, could be first name.

    Only can be confident in this if there are two tokens, and at least one of
    them is only a single letter (optionally followed by period).

    Two capital letters in the first name is not sufficient, for example TJ
    is probably the first name.

    Also looks at if middle_name is already filled.

    Misses some cases that we could worry about, but this gets the vast majority.
    """
    idx = first.notnull() & middle.isnull()

    # if pd.notna(name.middle_name) or pd.isna(name.first_name):
    #     return name
    pattern = r"^\s*(\w+)\.?\s+(\w+)\.?\s*$"
    a = first.re_extract(pattern, 1).nullif("")
    b = first.re_extract(pattern, 2).nullif("")

    # Deal with "Kay Ellen", "E" should yield "Kay", "Ellen"
    middle_is_middle = (starts_with(middle, b) | starts_with(b, middle)).fill_null(
        False
    )
    result_first = middle_is_middle.ifelse(a, first)
    result_middle = middle_is_middle.ifelse(choose_longer(b, middle), middle)

    al = a.length().fill_null(0)
    bl = b.length().fill_null(0)
    short_long = (al == 1) & (bl > 1)  # A Jones
    long_short = (al > 1) & (bl == 1)  # Alice J
    idx &= short_long | long_short

    # Many rows are of the form first_name="H Daniel", last_name="Hull"
    # where the first token of the first name is actually the
    # first letter of the last name. Catch this.
    first_is_last = starts_with(last, a).fill_null(False)
    fil = idx & first_is_last
    result_first = fil.ifelse(b, result_first)
    result_middle = fil.ifelse(ibis.null(), result_middle)

    finl = idx & ~first_is_last
    result_first = finl.ifelse(a, result_first)
    result_middle = finl.ifelse(b, result_middle)

    # Correct for when the last name is "A Jones"
    a = last.re_extract(pattern, 1).nullif("")
    b = last.re_extract(pattern, 2).nullif("")
    al = a.length().fill_null(0)
    bl = b.length().fill_null(0)
    idx = (al == 1) & (bl > 1)  # A Jones
    idx &= middle.isnull()
    result_middle = idx.ifelse(a, result_middle)
    result_last = idx.ifelse(b, last)

    return result_first, result_middle, result_last


def fix_middle(names: ir.Table) -> ir.Table:
    f, m, la = parse_middle(
        names["first_name"], names["middle_name"], names["last_name"]
    )
    return names.mutate(first_name=f, middle_name=m, last_name=la)


def fix_nickname_is_middle(t: ir.Table) -> ir.Table:
    """Fix 'george "mike" m smith", Where mike is both the nickname and middle name.

    Watch out for when the nickname is probably not related to the middle name,
    Such as with 'Carolyn "Care" c smith' (Care is short for Carolyn, not the middle)
    """
    todo = starts_with(t["nickname"], t["middle_name"]).fill_null(False)
    # Get rid of the 'Carolyn "Care" c smith' case
    todo &= ~starts_with(t["first_name"], t["middle_name"])
    return t.mutate(middle_name=todo.ifelse(t.nickname, t.middle_name))


def fix_two_token_people(t: ir.Table) -> ir.Table:
    """If no first name, and a two-token last name, extract first and last into respective columns."""
    first, last = _extract_two_token_people(t.first_name, t.last_name)
    return t.mutate(first_name=first, last_name=last)


def _extract_two_token_people(
    first: ir.StringColumn, last: ir.StringColumn
) -> tuple[ir.StringColumn, ir.StringColumn]:
    first = first.nullif("")
    last = last.nullif("")
    tokens = last.split(" ")
    fi = norm_whitespace(tokens[0])
    la = norm_whitespace(tokens[1])
    idx = last.notnull() & first.isnull() & (tokens.length() == 2)
    return idx.ifelse(fi, first), idx.ifelse(la, last)


def fix_last_comma_first(t: ir.Table) -> ir.Table:
    """Parses when first name is NA, and last name follows the form "Last, First"."""
    pattern = r"^(.*),(.*)$"
    a = t.last_name.re_extract(pattern, 1)
    b = t.last_name.re_extract(pattern, 2)
    a = norm_whitespace(a)
    b = norm_whitespace(b)
    one_each = (num_tokens(a) == 1) & (num_tokens(b) == 1)
    first_empty = t.first_name.strip().fill_null("") == ""
    todo = first_empty & one_each
    return t.mutate(
        first_name=todo.ifelse(b, t.first_name),
        last_name=todo.ifelse(a, t.last_name),
    )


def norm_name_fields(t: ir.Table) -> ir.Table:
    """Normalize capitalization and whitespace, remove weird punctuation, etc."""
    t = t.mutate(**{col: norm_name_field(t[col]) for col in NAME_COLUMNS})
    # suffixes have some specialcapitalization rules
    t = t.mutate(prefix=_norm_affix(t.prefix), suffix=_norm_affix(t.suffix))
    return t


def norm_name_field(s: ir.StringValue) -> ir.StringValue:
    """Fix capitalization and whitespace in a name field.

    "DEBOER" -> "Deboer"
    "DeBoer" -> "DeBoer"
    "OBrien" -> "OBrien"
    "O'Brien" -> "O'Brien"
    "Jr. " -> "Jr"
    "Abc -Def" -> "Abc-Def"
    """

    def _norm_token(t: ir.StringValue) -> ir.StringValue:
        num_upper = t.re_replace(r"[^A-Z]", "").length()
        num_lower = t.re_replace(r"[^a-z]", "").length()
        multi_caps = (num_upper >= 2) & (num_lower >= 1)
        caps_in_middler = (num_upper >= 1) & t.re_search(r"[a-z]")
        purposefully_weird = multi_caps | caps_in_middler
        return purposefully_weird.ifelse(t, t.capitalize())

    s = s.re_replace(r"[^a-zA-Z \-']", "")
    s = s.split(" ").map(_norm_token).join(" ")
    s = s.re_replace(r"\s+", " ")
    s = s.re_replace(r"(\W) (\W)", r"\1\2")  # between non-words
    s = s.re_replace(r"(\w) (\W)", r"\1\2")  # between word and non-word
    s = s.re_replace(r"(\W) (\w)", r"\1\2")  # between non-word and word
    s = s.strip()
    return s


def _norm_affix(s: ir.StringValue) -> ir.StringValue:
    s = s.upper()
    s = s.substitute(
        {
            "JR": "Jr",
            "SR": "Sr",
            "PHD": "PhD",
            "MR": "Mr",
            "MRS": "Mrs",
            "MS": "Ms",
            "DR": "Dr",
        }
    )
    return s


def clean_names(names: ir.Table) -> ir.Table:
    """Perform all the cleaning steps on a Table of names."""
    names = names.mutate(**{col: normalize_nulls(names[col]) for col in NAME_COLUMNS})
    # names = names.cache()
    names = fix_nicknames(names)
    # names = names.cache()
    names = fix_last_comma_first(names)
    # names = names.cache()
    names = fix_two_token_people(names)
    # names = names.cache()
    names = fix_middle(names)
    # names = names.cache()
    names = drop_first_when_same_as_last(names)
    # names = names.cache()
    names = fix_duplicate_appearances(names)
    # names = names.cache()
    names = norm_name_fields(names)
    # names = names.cache()
    names = fix_nickname_is_middle(names)
    # names = names.cache()
    return names



from __future__ import annotations

import contextlib
import os
import webbrowser
from typing import TYPE_CHECKING, Any, NoReturn

from public import public

import ibis
import ibis.expr.operations as ops
from ibis.common.annotations import ValidationError
from ibis.common.exceptions import IbisError, TranslationError
from ibis.common.grounds import Immutable
from ibis.common.patterns import Coercible, CoercionError
from ibis.common.typing import get_defining_scope
from ibis.config import _default_backend
from ibis.config import options as opts
from ibis.expr.format import pretty
from ibis.util import experimental

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from pathlib import Path

    import pandas as pd
    import polars as pl
    import pyarrow as pa
    import torch
    from rich.console import Console, RenderableType

    import ibis.expr.types as ir
    from ibis.backends import BaseBackend
    from ibis.expr.visualize import EdgeAttributeGetter, NodeAttributeGetter


try:
    from rich.jupyter import JupyterMixin
except ImportError:

    class _FixedTextJupyterMixin:
        """No-op when rich is not installed."""
else:

    class _FixedTextJupyterMixin(JupyterMixin):
        """JupyterMixin adds a spurious newline to text, this fixes the issue."""

        def _repr_mimebundle_(self, *args, **kwargs):
            try:
                bundle = super()._repr_mimebundle_(*args, **kwargs)
            except Exception:  # noqa: BLE001
                return None
            else:
                bundle["text/plain"] = bundle["text/plain"].rstrip()
                return bundle


def _capture_rich_renderable(renderable: RenderableType) -> str:
    from rich.console import Console

    console = Console(force_terminal=False)
    with console.capture() as capture:
        console.print(renderable)
    return capture.get().rstrip()


@public
class Expr(Immutable, Coercible):
    """Base expression class."""

    __slots__ = ("_arg",)
    _arg: ops.Node

    def _noninteractive_repr(self) -> str:
        if ibis.options.repr.show_variables:
            scope = get_defining_scope(self, types=Expr)
        else:
            scope = None
        return pretty(self.op(), scope=scope)

    def __repr__(self) -> str:
        if ibis.options.interactive:
            return _capture_rich_renderable(self)
        else:
            return self._noninteractive_repr()

    def __rich_console__(self, console: Console, options):
        from rich.text import Text

        if console.is_jupyter:
            # Rich infers a console width in jupyter notebooks, but since
            # notebooks can use horizontal scroll bars we don't want to apply a
            # limit here. Since rich requires an integer for max_width, we
            # choose an arbitrarily large integer bound. Note that we need to
            # handle this here rather than in `to_rich`, as this setting
            # also needs to be forwarded to `console.render`.
            options = options.update(max_width=1_000_000)
            console_width = None
        else:
            console_width = options.max_width

        try:
            if opts.interactive:
                from ibis.expr.types.pretty import to_rich

                rich_object = to_rich(self, console_width=console_width)
            else:
                rich_object = Text(self._noninteractive_repr())
        except TranslationError as e:
            lines = [
                "Translation to backend failed",
                f"Error message: {e!r}",
                "Expression repr follows:",
                self._noninteractive_repr(),
            ]
            return Text("\n".join(lines))
        return console.render(rich_object, options=options)

    def __init__(self, arg: ops.Node) -> None:
        object.__setattr__(self, "_arg", arg)

    def __iter__(self) -> NoReturn:
        raise TypeError(f"{self.__class__.__name__!r} object is not iterable")

    @classmethod
    def __coerce__(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, ops.Node):
            return value.to_expr()
        else:
            raise CoercionError("Unable to coerce value to an expression")

    def __reduce__(self):
        return (self.__class__, (self._arg,))

    def __hash__(self):
        return hash((self.__class__, self._arg))

    def equals(self, other) -> bool:
        """Return whether this expression is _structurally_ equivalent to `other`.

        If you want to produce an equality expression, use `==` syntax.

        Parameters
        ----------
        other
            Another expression

        Examples
        --------
        >>> import ibis
        >>> t1 = ibis.table(dict(a="int"), name="t")
        >>> t2 = ibis.table(dict(a="int"), name="t")
        >>> t1.equals(t2)
        True
        >>> v = ibis.table(dict(a="string"), name="v")
        >>> t1.equals(v)
        False
        """
        if not isinstance(other, Expr):
            raise TypeError(
                f"invalid equality comparison between Expr and {type(other)}"
            )
        return self._arg.equals(other._arg)

    def __bool__(self) -> bool:
        raise ValueError("The truth value of an Ibis expression is not defined")

    __nonzero__ = __bool__

    def get_name(self):
        """Return the name of this expression."""
        return self._arg.name

    def _repr_png_(self) -> bytes | None:
        if opts.interactive or not opts.graphviz_repr:
            return None
        try:
            import ibis.expr.visualize as viz
        except ImportError:
            return None
        else:
            # Something may go wrong, and we can't error in the notebook
            # so fallback to the default text representation.
            with contextlib.suppress(Exception):
                return viz.to_graph(self).pipe(format="png")

    def visualize(
        self,
        format: str = "svg",
        *,
        label_edges: bool = False,
        verbose: bool = False,
        node_attr: Mapping[str, str] | None = None,
        node_attr_getter: NodeAttributeGetter | None = None,
        edge_attr: Mapping[str, str] | None = None,
        edge_attr_getter: EdgeAttributeGetter | None = None,
    ) -> None:
        """Visualize an expression as a GraphViz graph in the browser.

        Parameters
        ----------
        format
            Image output format. These are specified by the `graphviz` Python
            library.
        label_edges
            Show operation input names as edge labels
        verbose
            Print the graphviz DOT code to stderr if [](`True`)
        node_attr
            Mapping of `(attribute, value)` pairs set for all nodes.
            Options are specified by the `graphviz` Python library.
        node_attr_getter
            Callback taking a node and returning a mapping of `(attribute, value)` pairs
            for that node. Options are specified by the `graphviz` Python library.
        edge_attr
            Mapping of `(attribute, value)` pairs set for all edges.
            Options are specified by the `graphviz` Python library.
        edge_attr_getter
            Callback taking two adjacent nodes and returning a mapping of `(attribute, value)` pairs
            for the edge between those nodes. Options are specified by the `graphviz` Python library.

        Examples
        --------
        Open the visualization of an expression in default browser:

        >>> import ibis
        >>> import ibis.expr.operations as ops
        >>> left = ibis.table(dict(a="int64", b="string"), name="left")
        >>> right = ibis.table(dict(b="string", c="int64", d="string"), name="right")
        >>> expr = left.inner_join(right, "b").select(left.a, b=right.c, c=right.d)
        >>> expr.visualize(
        ...     format="svg",
        ...     label_edges=True,
        ...     node_attr={"fontname": "Roboto Mono", "fontsize": "10"},
        ...     node_attr_getter=lambda node: isinstance(node, ops.Field) and {"shape": "oval"},
        ...     edge_attr={"fontsize": "8"},
        ...     edge_attr_getter=lambda u, v: isinstance(u, ops.Field) and {"color": "red"},
        ... )  # quartodoc: +SKIP # doctest: +SKIP

        Raises
        ------
        ImportError
            If `graphviz` is not installed.
        """
        import ibis.expr.visualize as viz

        path = viz.draw(
            viz.to_graph(
                self,
                node_attr=node_attr,
                node_attr_getter=node_attr_getter,
                edge_attr=edge_attr,
                edge_attr_getter=edge_attr_getter,
                label_edges=label_edges,
            ),
            format=format,
            verbose=verbose,
        )
        webbrowser.open(f"file://{os.path.abspath(path)}")

    def pipe(self, f, *args: Any, **kwargs: Any) -> Expr:
        """Compose `f` with `self`.

        Parameters
        ----------
        f
            If the expression needs to be passed as anything other than the
            first argument to the function, pass a tuple with the argument
            name. For example, (f, 'data') if the function f expects a 'data'
            keyword
        args
            Positional arguments to `f`
        kwargs
            Keyword arguments to `f`

        Examples
        --------
        >>> import ibis
        >>> t = ibis.memtable(
        ...     {
        ...         "a": [5, 10, 15],
        ...         "b": ["a", "b", "c"],
        ...     }
        ... )
        >>> f = lambda a: (a + 1).name("a")
        >>> g = lambda a: (a * 2).name("a")
        >>> result1 = t.a.pipe(f).pipe(g)
        >>> result1
        r0 := InMemoryTable
        data:
            PandasDataFrameProxy:
                a  b
            0   5  a
            1  10  b
            2  15  c
        a: r0.a + 1 * 2

        >>> result2 = g(f(t.a))  # equivalent to the above
        >>> result1.equals(result2)
        True

        Returns
        -------
        Expr
            Result type of passed function
        """
        if isinstance(f, tuple):
            f, data_keyword = f
            kwargs = kwargs.copy()
            kwargs[data_keyword] = self
            return f(*args, **kwargs)
        else:
            return f(self, *args, **kwargs)

    def op(self) -> ops.Node:
        return self._arg

    def _find_backends(self) -> tuple[list[BaseBackend], bool]:
        """Return the possible backends for an expression.

        Returns
        -------
        list[BaseBackend]
            A list of the backends found.
        """

        backends = set()
        has_unbound = False
        node_types = (ops.UnboundTable, ops.DatabaseTable, ops.SQLQueryResult)
        for table in self.op().find(node_types):
            if isinstance(table, ops.UnboundTable):
                has_unbound = True
            else:
                backends.add(table.source)

        return list(backends), has_unbound

    def _find_backend(self, *, use_default: bool = False) -> BaseBackend:
        """Find the backend attached to an expression.

        Parameters
        ----------
        use_default
            If [](`True`) and the default backend isn't set, initialize the
            default backend and use that. This should only be set to `True` for
            `.execute()`. For other contexts such as compilation, this option
            doesn't make sense so the default value is [](`False`).

        Returns
        -------
        BaseBackend
            A backend that is attached to the expression
        """
        backends, has_unbound = self._find_backends()

        if not backends:
            if has_unbound:
                raise IbisError(
                    "Expression contains unbound tables and therefore cannot "
                    "be executed. Use `<backend>.execute(expr)` to execute "
                    "against an explicit backend, or rebuild the expression "
                    "using bound tables instead."
                )
            default = _default_backend() if use_default else None
            if default is None:
                raise IbisError(
                    "Expression depends on no backends, and found no default"
                )
            return default

        if len(backends) > 1:
            raise IbisError("Multiple backends found for this expression")

        return backends[0]

    def get_backend(self) -> BaseBackend:
        """Get the current Ibis backend of the expression.

        Returns
        -------
        BaseBackend
            The Ibis backend.

        Examples
        --------
        >>> import ibis
        >>> con = ibis.duckdb.connect()
        >>> t = con.create_table("t", {"id": [1, 2, 3]})
        >>> t.get_backend()  # doctest: +ELLIPSIS
        <ibis.backends.duckdb.Backend object at 0x...>

        See Also
        --------
        [`ibis.get_backend()`](./connection.qmd#ibis.get_backend)
        """
        return self._find_backend(use_default=True)

    def execute(
        self,
        limit: int | str | None = "default",
        params: Mapping[ir.Value, Any] | None = None,
        **kwargs: Any,
    ):
        """Execute an expression against its backend if one exists.

        Parameters
        ----------
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        params
            Mapping of scalar parameter expressions to value
        kwargs
            Keyword arguments

        Examples
        --------
        >>> import ibis
        >>> t = ibis.examples.penguins.fetch()
        >>> t.execute()
               species     island  bill_length_mm  ...  body_mass_g     sex  year
        0       Adelie  Torgersen            39.1  ...       3750.0    male  2007
        1       Adelie  Torgersen            39.5  ...       3800.0  female  2007
        2       Adelie  Torgersen            40.3  ...       3250.0  female  2007
        3       Adelie  Torgersen             NaN  ...          NaN    None  2007
        4       Adelie  Torgersen            36.7  ...       3450.0  female  2007
        ..         ...        ...             ...  ...          ...     ...   ...
        339  Chinstrap      Dream            55.8  ...       4000.0    male  2009
        340  Chinstrap      Dream            43.5  ...       3400.0  female  2009
        341  Chinstrap      Dream            49.6  ...       3775.0    male  2009
        342  Chinstrap      Dream            50.8  ...       4100.0    male  2009
        343  Chinstrap      Dream            50.2  ...       3775.0  female  2009
        [344 rows x 8 columns]

        Scalar parameters can be supplied dynamically during execution.
        >>> species = ibis.param("string")
        >>> expr = t.filter(t.species == species).order_by(t.bill_length_mm)
        >>> expr.execute(limit=3, params={species: "Gentoo"})
          species  island  bill_length_mm  ...  body_mass_g     sex  year
        0  Gentoo  Biscoe            40.9  ...         4650  female  2007
        1  Gentoo  Biscoe            41.7  ...         4700  female  2009
        2  Gentoo  Biscoe            42.0  ...         4150  female  2007
        <BLANKLINE>
        [3 rows x 8 columns]

        See Also
        --------
        [`Table.to_pandas()`](./expression-tables.qmd#ibis.expr.types.relations.Table.to_pandas)
        [`Value.to_pandas()`](./expression-generic.qmd#ibis.expr.types.generic.Value.to_pandas)
        """
        return self._find_backend(use_default=True).execute(
            self, limit=limit, params=params, **kwargs
        )

    def compile(
        self,
        limit: int | None = None,
        params: Mapping[ir.Value, Any] | None = None,
        pretty: bool = False,
    ) -> str:
        r"""Compile to an execution target.

        Parameters
        ----------
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        params
            Mapping of scalar parameter expressions to value
        pretty
            In case of SQL backends, return a pretty formatted SQL query.

        Returns
        -------
        str
            SQL query representation of the expression

        Examples
        --------
        >>> import ibis
        >>> d = {"a": [1, 2, 3], "b": [4, 5, 6]}
        >>> con = ibis.duckdb.connect()
        >>> t = con.create_table("t", d)
        >>> expr = t.mutate(c=t.a + t.b)
        >>> expr.compile()
        'SELECT "t0"."a", "t0"."b", "t0"."a" + "t0"."b" AS "c" FROM "memory"."main"."t" AS "t0"'

        If you want to see the pretty formatted SQL query, set `pretty` to `True`.
        >>> expr.compile(pretty=True)
        'SELECT\n  "t0"."a",\n  "t0"."b",\n  "t0"."a" + "t0"."b" AS "c"\nFROM "memory"."main"."t" AS "t0"'

        If the expression does not have a backend, an error will be raised.
        >>> t = ibis.memtable(d)
        >>> expr = t.mutate(c=t.a + t.b)
        >>> expr.compile()  # quartodoc: +EXPECTED_FAILURE
        Traceback (most recent call last):
        ...
        ibis.common.exceptions.IbisError: Expression depends on no backends, and ...

        See Also
        --------
        [`ibis.to_sql()`](./expression-generic.qmd#ibis.to_sql)
        """
        return self._find_backend().compile(
            self, limit=limit, params=params, pretty=pretty
        )

    @experimental
    def to_pyarrow_batches(
        self,
        *,
        limit: int | str | None = None,
        params: Mapping[ir.Value, Any] | None = None,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        """Execute expression and return a RecordBatchReader.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        params
            Mapping of scalar parameter expressions to value.
        chunk_size
            Maximum number of rows in each returned record batch.
        kwargs
            Keyword arguments

        Returns
        -------
        results
            RecordBatchReader
        """
        return self._find_backend(use_default=True).to_pyarrow_batches(
            self,
            params=params,
            limit=limit,
            chunk_size=chunk_size,
            **kwargs,
        )

    @experimental
    def to_pyarrow(
        self,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        """Execute expression and return results in as a pyarrow table.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        kwargs
            Keyword arguments

        Returns
        -------
        Table
            A pyarrow table holding the results of the executed expression.
        """
        return self._find_backend(use_default=True).to_pyarrow(
            self, params=params, limit=limit, **kwargs
        )

    @experimental
    def to_polars(
        self,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pl.DataFrame:
        """Execute expression and return results as a polars dataframe.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        kwargs
            Keyword arguments

        Returns
        -------
        DataFrame
            A polars dataframe holding the results of the executed expression.
        """
        return self._find_backend(use_default=True).to_polars(
            self, params=params, limit=limit, **kwargs
        )

    @experimental
    def to_pandas_batches(
        self,
        *,
        limit: int | str | None = None,
        params: Mapping[ir.Value, Any] | None = None,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ) -> Iterator[pd.DataFrame | pd.Series | Any]:
        """Execute expression and return an iterator of pandas DataFrames.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        params
            Mapping of scalar parameter expressions to value.
        chunk_size
            Maximum number of rows in each returned `DataFrame`.
        kwargs
            Keyword arguments

        Returns
        -------
        Iterator[pd.DataFrame]
        """
        return self._find_backend(use_default=True).to_pandas_batches(
            self,
            params=params,
            limit=limit,
            chunk_size=chunk_size,
            **kwargs,
        )

    @experimental
    def to_parquet(
        self,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a parquet file.

        This method is eager and will execute the associated expression
        immediately.

        See https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetWriter.html for details.

        Parameters
        ----------
        path
            The data source. A string or Path to the parquet file.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            Additional keyword arguments passed to pyarrow.parquet.ParquetWriter

        Examples
        --------
        Write out an expression to a single parquet file.

        >>> import ibis
        >>> import tempfile
        >>> penguins = ibis.examples.penguins.fetch()
        >>> penguins.to_parquet(tempfile.mktemp())

        Partition on a single column.

        >>> penguins.to_parquet(tempfile.mkdtemp(), partition_by="year")

        Partition on multiple columns.

        >>> penguins.to_parquet(tempfile.mkdtemp(), partition_by=("year", "island"))

        ::: {.callout-note}
        ## Hive-partitioned output is currently only supported when using DuckDB
        :::
        """
        self._find_backend(use_default=True).to_parquet(self, path, **kwargs)

    @experimental
    def to_parquet_dir(
        self,
        directory: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a parquet file in a directory.

        This method is eager and will execute the associated expression
        immediately.

        See https://arrow.apache.org/docs/python/generated/pyarrow.dataset.write_dataset.html for details.

        Parameters
        ----------
        directory
            The data target. A string or Path to the directory where the parquet file will be written.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            Additional keyword arguments passed to pyarrow.dataset.write_dataset
        """
        self._find_backend(use_default=True).to_parquet_dir(self, directory, **kwargs)

    @experimental
    def to_csv(
        self,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a CSV file.

        This method is eager and will execute the associated expression
        immediately.

        See https://arrow.apache.org/docs/python/generated/pyarrow.csv.CSVWriter.html for details.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            Additional keyword arguments passed to pyarrow.csv.CSVWriter
        """
        self._find_backend(use_default=True).to_csv(self, path, **kwargs)

    @experimental
    def to_delta(
        self,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a Delta Lake table.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        path
            The data source. A string or Path to the Delta Lake table directory.
        params
            Mapping of scalar parameter expressions to value.
        **kwargs
            Additional keyword arguments passed to deltalake.writer.write_deltalake method
        """
        self._find_backend(use_default=True).to_delta(self, path, **kwargs)

    @experimental
    def to_torch(
        self,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> dict[str, torch.Tensor]:
        """Execute an expression and return results as a dictionary of torch tensors.

        Parameters
        ----------
        params
            Parameters to substitute into the expression.
        limit
            An integer to effect a specific row limit. A value of `None` means no limit.
        kwargs
            Keyword arguments passed into the backend's `to_torch` implementation.

        Returns
        -------
        dict[str, torch.Tensor]
            A dictionary of torch tensors, keyed by column name.
        """
        return self._find_backend(use_default=True).to_torch(
            self, params=params, limit=limit, **kwargs
        )

    def unbind(self) -> ir.Table:
        """Return an expression built on `UnboundTable` instead of backend-specific objects.

        Examples
        --------
        >>> import ibis
        >>> import pandas as pd
        >>> duckdb_con = ibis.duckdb.connect()
        >>> polars_con = ibis.polars.connect()
        >>> for backend in (duckdb_con, polars_con):
        ...     t = backend.create_table("t", pd.DataFrame({"a": [1, 2, 3]}))
        >>> bound_table = duckdb_con.table("t")
        >>> bound_table.get_backend().name
        'duckdb'
        >>> unbound_table = bound_table.unbind()
        >>> polars_con.execute(unbound_table)
           a
        0  1
        1  2
        2  3
        """
        from ibis.expr.rewrites import _, d, p

        rule = p.DatabaseTable >> d.UnboundTable(
            name=_.name, schema=_.schema, namespace=_.namespace
        )
        return self.op().replace(rule).to_expr()

    def as_table(self) -> ir.Table:
        """Convert an expression to a table."""
        raise NotImplementedError(
            f"{type(self)} expressions cannot be converted into tables"
        )

    def as_scalar(self) -> ir.Scalar:
        """Convert an expression to a scalar."""
        raise NotImplementedError(
            f"{type(self)} expressions cannot be converted into scalars"
        )


def _binop(op_class: type[ops.Binary], left: ir.Value, right: ir.Value) -> ir.Value:
    """Try to construct a binary operation.

    Parameters
    ----------
    op_class
        The `ops.Binary` subclass for the operation
    left
        Left operand
    right
        Right operand

    Returns
    -------
    ir.Value
        A value expression

    Examples
    --------
    >>> import ibis
    >>> import ibis.expr.operations as ops
    >>> expr = _binop(ops.TimeAdd, ibis.time("01:00"), ibis.interval(hours=1))
    >>> expr
    TimeAdd(datetime.time(1, 0), 1h): datetime.time(1, 0) + 1 h
    >>> _binop(ops.TimeAdd, 1, ibis.interval(hours=1))
    TimeAdd(datetime.time(0, 0, 1), 1h): datetime.time(0, 0, 1) + 1 h
    """
    try:
        node = op_class(left, right)
    except (ValidationError, NotImplementedError):
        return NotImplemented
    else:
        return node.to_expr()


def _is_null_literal(value: Any) -> bool:
    """Detect whether `value` will be treated by ibis as a null literal."""
    if value is None:
        return True
    if isinstance(value, Expr):
        op = value.op()
        return isinstance(op, ops.Literal) and op.value is None
    return False
